@@ -15,6 +15,7 @@ use std::{
 };
 // reexport the lazy method
 pub use crate::frame::IntoLazy;
+use polars_core::frame::select::Selection;
 use polars_core::utils::get_supertype;
 
 /// A wrapper trait for any closure `Fn(Vec<Series>) -> Result<Series>`
@@ -260,7 +261,9 @@ pub enum Expr {
         output_field: NoEq<Arc<dyn BinaryUdfOutputField>>,
     },
     /// Can be used in a select statement to exclude a column from selection
-    Except(Box<Expr>),
+    Exclude(Box<Expr>, Vec<Arc<String>>),
+    /// Set root name as Alias
+    KeepName(Box<Expr>),
 }
 
 impl Expr {
@@ -351,34 +354,12 @@ impl fmt::Debug for Expr {
                 length,
             } => write!(f, "SLICE {:?} offset: {} len: {}", input, offset, length),
             Wildcard => write!(f, "*"),
-            Except(column) => write!(f, "EXCEPT {:?}", column),
+            Exclude(column, names) => write!(f, "{:?}, EXCEPT {:?}", column, names),
+            KeepName(e) => write!(f, "KEEP NAME {:?}", e),
         }
     }
 }
 
-/// Exclude a column from selection.
-///
-/// # Example
-///
-/// ```rust
-/// use polars_core::prelude::*;
-/// use polars_lazy::prelude::*;
-///
-/// // Select all columns except foo.
-/// fn example(df: DataFrame) -> LazyFrame {
-///       df.lazy()
-///         .select(&[
-///                 col("*"), except("foo")
-///                 ])
-/// }
-/// ```
-pub fn except(name: &str) -> Expr {
-    match name {
-        "*" => panic!("cannot use a wildcard as a column exception"),
-        _ => Expr::Except(Box::new(col(name))),
-    }
-}
-
 #[derive(Debug, Copy, Clone, PartialEq)]
 pub enum Operator {
     Eq,
@@ -1146,12 +1127,81 @@ impl Expr {
 
     #[cfg(feature = "mode")]
     #[cfg_attr(docsrs, doc(cfg(feature = "mode")))]
+    /// Compute the mode(s) of this column. These is the most occurring value.
     pub fn mode(self) -> Expr {
         self.map(|s| s.mode().map(|ca| ca.into_series()), None)
     }
+
+    /// Keep the original root name
+    ///
+    /// ```
+    /// use polars_core::prelude::*;
+    /// use polars_lazy::prelude::*;
+    ///
+    /// fn example(df: LazyFrame) -> LazyFrame {
+    ///     df.select(vec![
+    /// // even thought the alias yields a different column name,
+    /// // `keep_name` will make sure that the original column name is used
+    ///         col("*").alias("foo").keep_name()
+    /// ])
+    /// }
+    /// ```
+    pub fn keep_name(self) -> Expr {
+        Expr::KeepName(Box::new(self))
+    }
+
+    /// Exclude a column from a wildcard selection
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// use polars_core::prelude::*;
+    /// use polars_lazy::prelude::*;
+    ///
+    /// // Select all columns except foo.
+    /// fn example(df: DataFrame) -> LazyFrame {
+    ///       df.lazy()
+    ///         .select(&[
+    ///                 col("*").exclude(&["foo"])
+    ///                 ])
+    /// }
+    /// ```
+    pub fn exclude<'a, S, J>(self, columns: S) -> Expr
+    where
+        S: Selection<'a, J>,
+    {
+        let v = columns
+            .to_selection_vec()
+            .iter()
+            .map(|s| Arc::new(s.to_string()))
+            .collect();
+        Expr::Exclude(Box::new(self), v)
+    }
 }
 
 /// Create a Column Expression based on a column name.
+///
+/// # Arguments
+///
+/// * `name` - A string slice that holds the name of the column
+///
+/// # Examples
+///
+/// ```ignore
+/// // select a column name
+/// col("foo")
+/// ```
+///
+/// ```ignore
+/// // select all columns by using a wildcard
+/// col("*")
+/// ```
+///
+/// ```ignore
+/// // select specific column by writing a regular expression that starts with `^` and ends with `$`
+/// // only if regex features is activated
+/// col("^foo.*$")
+/// ```
 pub fn col(name: &str) -> Expr {
     match name {
         "*" => Expr::Wildcard,
