@@ -22,10 +22,11 @@ use polars_io::{parquet::ParquetReader, SerReader};
 
 use crate::logical_plan::LogicalPlan::DataFrameScan;
 use crate::utils::{
-    combine_predicates_expr, expr_to_root_column_name, expr_to_root_column_names, has_expr,
+    combine_predicates_expr, expr_to_root_column_names, has_expr, has_wildcard,
     rename_expr_root_name,
 };
 use crate::{prelude::*, utils};
+use polars_io::csv::NullValues;
 
 pub(crate) mod aexpr;
 pub(crate) mod alp;
@@ -146,6 +147,7 @@ pub struct CsvParserOptions {
     pub(crate) low_memory: bool,
     pub(crate) ignore_errors: bool,
     pub(crate) cache: bool,
+    pub(crate) null_values: Option<NullValues>,
 }
 
 // https://stackoverflow.com/questions/1031076/what-are-projection-and-selection
@@ -668,163 +670,75 @@ impl LogicalPlan {
     }
 }
 
-fn replace_wildcard_with_column(expr: Expr, column_name: Arc<String>) -> Expr {
-    match expr {
-        Expr::Window {
-            function,
-            partition_by,
-            order_by,
-        } => Expr::Window {
-            function: Box::new(replace_wildcard_with_column(*function, column_name)),
-            partition_by,
-            order_by,
-        },
-        Expr::IsUnique(expr) => {
-            Expr::IsUnique(Box::new(replace_wildcard_with_column(*expr, column_name)))
+/// This replace the wilcard Expr with a Column Expr. It also removes the Exclude Expr from the
+/// expression chain.
+fn replace_wildcard_with_column(mut expr: Expr, column_name: Arc<String>) -> Expr {
+    expr.mutate().apply(|e| {
+        match &e {
+            Expr::Wildcard => {
+                *e = Expr::Column(column_name.clone());
+            }
+            Expr::Exclude(input, _) => {
+                *e = replace_wildcard_with_column(*input.clone(), column_name.clone());
+            }
+            _ => {}
         }
-        Expr::Duplicated(expr) => {
-            Expr::Duplicated(Box::new(replace_wildcard_with_column(*expr, column_name)))
+        // always keep iterating all inputs
+        true
+    });
+    expr
+}
+
+fn rewrite_keep_name(expr: Expr) -> Expr {
+    if has_expr(&expr, |e| matches!(e, Expr::KeepName(_))) {
+        if let Expr::KeepName(expr) = expr {
+            let roots = expr_to_root_column_names(&expr);
+            let name = roots
+                .get(0)
+                .expect("expected root column to keep expression name");
+            Expr::Alias(expr, name.clone())
+        } else {
+            panic!("keep_name should be last expression")
         }
-        Expr::Reverse(expr) => {
-            Expr::Reverse(Box::new(replace_wildcard_with_column(*expr, column_name)))
+    } else {
+        expr
+    }
+}
+
+/// Take an expression with a root: col("*") and copies that expression for all columns in the schema,
+/// with the exclusion of the `names` in the exclude expression.
+/// The resulting expressions are written to result.
+fn replace_wilcard(expr: &Expr, result: &mut Vec<Expr>, exclude: &[Arc<String>], schema: &Schema) {
+    for field in schema.fields() {
+        let name = field.name();
+        if !exclude.iter().any(|exluded| &**exluded == name) {
+            let new_expr = replace_wildcard_with_column(expr.clone(), Arc::new(name.clone()));
+            let new_expr = rewrite_keep_name(new_expr);
+            result.push(new_expr)
         }
-        Expr::Explode(expr) => {
-            Expr::Explode(Box::new(replace_wildcard_with_column(*expr, column_name)))
+    }
+}
+
+#[cfg(feature = "regex")]
+fn replace_regex(expr: &Expr, result: &mut Vec<Expr>, schema: &Schema, pattern: &str) {
+    let re = regex::Regex::new(pattern)
+        .unwrap_or_else(|_| panic!("invalid regular expression in column: {}", pattern));
+    for field in schema.fields() {
+        let name = field.name();
+        if re.is_match(name) {
+            let mut new_expr = expr.clone();
+
+            new_expr.mutate().apply(|e| match &e {
+                Expr::Column(_) => {
+                    *e = Expr::Column(Arc::new(name.clone()));
+                    false
+                }
+                _ => true,
+            });
+
+            let new_expr = rewrite_keep_name(new_expr);
+            result.push(new_expr)
         }
-        Expr::Take { expr, idx } => Expr::Take {
-            expr: Box::new(replace_wildcard_with_column(*expr, column_name)),
-            idx,
-        },
-        Expr::Ternary {
-            predicate,
-            truthy,
-            falsy,
-        } => Expr::Ternary {
-            predicate: Box::new(replace_wildcard_with_column(
-                *predicate,
-                column_name.clone(),
-            )),
-            truthy: Box::new(replace_wildcard_with_column(*truthy, column_name.clone())),
-            falsy: Box::new(replace_wildcard_with_column(*falsy, column_name)),
-        },
-        Expr::Function {
-            input,
-            function,
-            output_type,
-            collect_groups,
-        } => Expr::Function {
-            input: input
-                .into_iter()
-                .map(|e| replace_wildcard_with_column(e, column_name.clone()))
-                .collect(),
-            function,
-            output_type,
-            collect_groups,
-        },
-        Expr::BinaryFunction {
-            input_a,
-            input_b,
-            function,
-            output_field,
-        } => Expr::BinaryFunction {
-            input_a: Box::new(replace_wildcard_with_column(*input_a, column_name.clone())),
-            input_b: Box::new(replace_wildcard_with_column(*input_b, column_name)),
-            function,
-            output_field,
-        },
-        Expr::BinaryExpr { left, op, right } => Expr::BinaryExpr {
-            left: Box::new(replace_wildcard_with_column(*left, column_name.clone())),
-            op,
-            right: Box::new(replace_wildcard_with_column(*right, column_name)),
-        },
-        Expr::Wildcard => Expr::Column(column_name),
-        Expr::IsNotNull(e) => {
-            Expr::IsNotNull(Box::new(replace_wildcard_with_column(*e, column_name)))
-        }
-        Expr::IsNull(e) => Expr::IsNull(Box::new(replace_wildcard_with_column(*e, column_name))),
-        Expr::Not(e) => Expr::Not(Box::new(replace_wildcard_with_column(*e, column_name))),
-        Expr::Alias(e, name) => Expr::Alias(
-            Box::new(replace_wildcard_with_column(*e, column_name)),
-            name,
-        ),
-        Expr::Filter { .. } => {
-            panic!("Expression filter may not be used with wildcard, use LazyFrame::filter")
-        }
-        Expr::Agg(agg) => match agg {
-            AggExpr::Mean(e) => {
-                AggExpr::Mean(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::Median(e) => {
-                AggExpr::Median(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::Max(e) => {
-                AggExpr::Max(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::Min(e) => {
-                AggExpr::Min(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::Sum(e) => {
-                AggExpr::Sum(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::Count(e) => {
-                AggExpr::Count(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::Last(e) => {
-                AggExpr::Last(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::First(e) => {
-                AggExpr::First(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::NUnique(e) => {
-                AggExpr::NUnique(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::AggGroups(e) => {
-                AggExpr::AggGroups(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::Quantile { expr, quantile } => AggExpr::Quantile {
-                expr: Box::new(replace_wildcard_with_column(*expr, column_name)),
-                quantile,
-            },
-            AggExpr::List(e) => {
-                AggExpr::List(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::Var(e) => {
-                AggExpr::Var(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-            AggExpr::Std(e) => {
-                AggExpr::Std(Box::new(replace_wildcard_with_column(*e, column_name)))
-            }
-        }
-        .into(),
-        Expr::Shift { input, periods } => Expr::Shift {
-            input: Box::new(replace_wildcard_with_column(*input, column_name)),
-            periods,
-        },
-        Expr::Slice {
-            input,
-            offset,
-            length,
-        } => Expr::Slice {
-            input: Box::new(replace_wildcard_with_column(*input, column_name)),
-            offset,
-            length,
-        },
-        Expr::SortBy { expr, by, reverse } => Expr::SortBy {
-            expr: Box::new(replace_wildcard_with_column(*expr, column_name)),
-            by,
-            reverse,
-        },
-        Expr::Sort { expr, reverse } => Expr::Sort {
-            expr: Box::new(replace_wildcard_with_column(*expr, column_name)),
-            reverse,
-        },
-        Expr::Cast { expr, data_type } => Expr::Cast {
-            expr: Box::new(replace_wildcard_with_column(*expr, column_name)),
-            data_type,
-        },
-        Expr::Column(_) => expr,
-        Expr::Literal(_) => expr,
-        Expr::Except(_) => expr,
     }
 }
 
@@ -832,22 +746,17 @@ fn replace_wildcard_with_column(expr: Expr, column_name: Arc<String>) -> Expr {
 /// In other cases replace the wildcard with an expression with all columns
 fn rewrite_projections(exprs: Vec<Expr>, schema: &Schema) -> Vec<Expr> {
     let mut result = Vec::with_capacity(exprs.len() + schema.fields().len());
-    let mut exclude = vec![];
-    for expr in exprs {
-        // Columns that are excepted are later removed from the projection.
-        // This can be ergonomical in combination with a wildcard expression.
-        if let Expr::Except(column) = &expr {
-            if let Expr::Column(name) = &**column {
-                exclude.push(name.clone());
-                continue;
-            } else {
-                panic!("Except expression should have column name")
-            }
-        }
 
-        let has_wildcard = has_expr(&expr, |e| matches!(e, Expr::Wildcard));
+    for mut expr in exprs {
+        if has_wildcard(&expr) {
+            // keep track of column excluded from the wildcard
+            let mut exclude = vec![];
+            (&expr).into_iter().for_each(|e| {
+                if let Expr::Exclude(_, names) = e {
+                    exclude.extend_from_slice(names)
+                }
+            });
 
-        if has_wildcard {
             // if count wildcard. count one column
             if has_expr(&expr, |e| matches!(e, Expr::Agg(AggExpr::Count(_)))) {
                 let new_name = Arc::new(schema.field(0).unwrap().name().clone());
@@ -862,28 +771,58 @@ fn rewrite_projections(exprs: Vec<Expr>, schema: &Schema) -> Vec<Expr> {
 
                 continue;
             }
+            // this path prepares the wildcard as input for the Function Expr
+            if has_expr(
+                &expr,
+                |e| matches!(e, Expr::Function { input, options,  .. } if options.input_wildcard_expansion && input.iter().any(has_wildcard)),
+            ) {
+                expr.mutate().apply(|e| {
+                    if let Expr::Function { input, .. } = e {
+                        let mut new_inputs = Vec::with_capacity(input.len());
 
-            for field in schema.fields() {
-                let name = field.name();
-                let new_expr = replace_wildcard_with_column(expr.clone(), Arc::new(name.clone()));
-                result.push(new_expr)
-            }
-        } else {
-            result.push(expr)
-        };
-    }
-    if !exclude.is_empty() {
-        for name in exclude {
-            let idx = result
-                .iter()
-                .position(|expr| match expr_to_root_column_name(expr) {
-                    Ok(column_name) => column_name == name,
-                    Err(_) => false,
+                        input.iter_mut().for_each(|e| {
+                            if has_wildcard(e) {
+                                replace_wilcard(e, &mut new_inputs, &[], schema)
+                            } else {
+                                new_inputs.push(e.clone())
+                            }
+                        });
+
+                        *input = new_inputs;
+                        false
+                    } else {
+                        true
+                    }
                 });
-            if let Some(idx) = idx {
-                result.swap_remove(idx);
+                result.push(expr);
+                continue;
             }
-        }
+            replace_wilcard(&expr, &mut result, &exclude, schema);
+        } else {
+            #[cfg(feature = "regex")]
+            {
+                // only in simple expression (no binary expression)
+                // we patter match regex columns
+                let roots = expr_to_root_column_names(&expr);
+                if roots.len() == 1 {
+                    let name = &**roots[0];
+                    if name.starts_with('^') && name.ends_with('$') {
+                        replace_regex(&expr, &mut result, schema, name)
+                    } else {
+                        let expr = rewrite_keep_name(expr);
+                        result.push(expr)
+                    }
+                } else {
+                    let expr = rewrite_keep_name(expr);
+                    result.push(expr)
+                }
+            }
+            #[cfg(not(feature = "regex"))]
+            {
+                let expr = rewrite_keep_name(expr);
+                result.push(expr)
+            }
+        };
     }
     result
 }
@@ -891,7 +830,7 @@ fn rewrite_projections(exprs: Vec<Expr>, schema: &Schema) -> Vec<Expr> {
 pub struct LogicalPlanBuilder(LogicalPlan);
 
 impl LogicalPlan {
-    pub(crate) fn schema(&self) -> &Schema {
+    pub(crate) fn schema(&self) -> &SchemaRef {
         use LogicalPlan::*;
         match self {
             Cache { input } => input.schema(),
@@ -976,6 +915,7 @@ impl LogicalPlanBuilder {
         schema_overwrite: Option<&Schema>,
         low_memory: bool,
         comment_char: Option<u8>,
+        null_values: Option<NullValues>,
     ) -> Self {
         let path = path.into();
         let mut file = std::fs::File::open(&path).expect("could not open file");
@@ -1006,6 +946,7 @@ impl LogicalPlanBuilder {
                 low_memory,
                 cache,
                 comment_char,
+                null_values,
             },
             predicate: None,
             aggregate: vec![],
@@ -1021,7 +962,7 @@ impl LogicalPlanBuilder {
     }
 
     pub fn project(self, exprs: Vec<Expr>) -> Self {
-        let (exprs, schema) = prepare_projection(exprs, &self.0.schema());
+        let (exprs, schema) = prepare_projection(exprs, self.0.schema());
 
         // if len == 0, no projection has to be done. This is a select all operation.
         if !exprs.is_empty() {
@@ -1037,7 +978,7 @@ impl LogicalPlanBuilder {
     }
 
     pub fn project_local(self, exprs: Vec<Expr>) -> Self {
-        let (exprs, schema) = prepare_projection(exprs, &self.0.schema());
+        let (exprs, schema) = prepare_projection(exprs, self.0.schema());
         if !exprs.is_empty() {
             LogicalPlan::LocalProjection {
                 expr: exprs,
@@ -1157,7 +1098,19 @@ impl LogicalPlanBuilder {
         .into()
     }
 
-    pub fn explode(self, columns: Vec<String>) -> Self {
+    pub fn explode(self, columns: Vec<Expr>) -> Self {
+        let columns = rewrite_projections(columns, self.0.schema());
+        // columns to string
+        let columns = columns
+            .iter()
+            .map(|e| {
+                if let Expr::Column(name) = e {
+                    (**name).clone()
+                } else {
+                    panic!("expected column expression")
+                }
+            })
+            .collect();
         LogicalPlan::Explode {
             input: Box::new(self.0),
             columns,
